module auxiliary

use constants

!-----------------------------------------------------------------------------------------------!
!       Auxiliary module for various calculations for CCS/2L-CCS simulation.                    !
!                                                                                               !
! Contains subroutines/functions to:                                                            !
! 1) Calculate coherent state overlap <z1|z2>                                                   !
! 2) Solve a linear equation of the form LHS*D=RHS for amplitude D                              !
! 3) Calculate the norm of the wavefunction                                                     !
! 4) Calculate the ACF or CCF                                                                   !
! 5) Calculate the density matrix (for CCSB)                                                    !
! 6) Calculate 1-body reduced density matrix (for CCSB)                                         !
! 7) Calculate the particle number (for CCSB)                                                   !
! 8) Calculate the variance of the 1-body density matrix (for CCSB)                             !
! 9) Evaluate a hermite polynomial                                                              !
!10) Calculate a factorial                                                                      !
!11) Calculate the bc_W_mnop Dirac delta function matrix (for CCSB)                             !
!12) Calculate an array of hermite polynomial coefficients                                      !
!13) Calculate the product of 4 hermite polynomial coefficients (for CCSB)                      !
!-----------------------------------------------------------------------------------------------!

contains

!----------------------------------------------------------------------------------------------------!

subroutine calc_ovrlp(ovrlp,z)

  implicit none

  complex(kind=8),intent(inout)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(in)::z(:,:,:)            !coherent state array (dof,nbv,nconf)
  integer::i,j,k,l

      !$omp parallel shared(ovrlp,z) private(i,j,k,l)
      !$omp do collapse(4)
      do l=1,size(ovrlp,4) !nconf
        do k=1,size(ovrlp,3) !nconf
          do j=1,size(ovrlp,2) !nbv (=1 for CCS)
            do i=1,size(ovrlp,1) !nbv (=1 for CCS)
              ovrlp(i,j,k,l) = exp(sum(conjg(z(:,i,k))*z(:,j,l)-0.5d0*(z(:,i,k)*conjg(z(:,i,k))+z(:,j,l)*conjg(z(:,j,l)))))
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

  return

end subroutine calc_ovrlp

!----------------------------------------------------------------------------------------------------!

subroutine lineq(LHS,RHS,D) !Linear equation : LHS*D=RHS

  implicit none

  complex(kind=8),intent(in)::LHS(:,:)             !left hand side of linear equation, usually coherent state overlap (nbv,nbv for d; nconf,nconf for D)
  complex(kind=8),intent(inout)::RHS(:)            !RHS of linear equation to calc amplitudes (nbv for d; nconf for D)
  complex(kind=8),intent(inout)::D(:)              !amplitude (nconf/nbv)
  integer::N,NRHS,INFO                             !Parameters for zgesv routine
  integer,allocatable::IPIV(:)
  complex(kind=8),allocatable::LHS_temp(:,:)       !Temp placeholder for LHS, as zgesv will overwrite its contents

  N=size(RHS)
  NRHS=1
  INFO=0
  allocate(IPIV(N),LHS_temp(size(LHS,1),size(LHS,2)))
  LHS_temp(:,:)=LHS(:,:)

  call zgesv(N,NRHS,LHS_temp,N,IPIV,RHS,N,INFO)

  D=RHS     !zgesv routine assigns the answer, D to the RHS input array

  deallocate(IPIV,LHS_temp)

  return

end subroutine lineq
!----------------------------------------------------------------------------------------------------!

function calc_norm(z,d,bigD,s,ovrlp)

  implicit none

  complex(kind=8),intent(in)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(in)::bigD(:)          !multi config amplitude (nconf)
  real(kind=8),intent(in)::s(:,:)              !action (nvb,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8)::calc_norm
  integer::i,j,k,l

  calc_norm=(0.d0,0.d0)

  !$omp parallel shared(d,s,ovrlp,bigD) private(l,k,j,i) reduction(+:calc_norm)
  !$omp do collapse(4)
  do l=1,size(ovrlp,4) !nconf
    do k=1,size(ovrlp,3) !nconf
      do j=1,size(d,1) !nbv (=1 for CCS)
        do i=1,size(d,1) !nbv (=1 for CCS)
            calc_norm = calc_norm + conjg(d(i,k))*d(j,l)*exp(im*(s(j,l)-s(i,k)))*ovrlp(i,j,k,l)*conjg(bigD(k))*bigD(l)
        end do
      end do
    end do
  end do
  !$omp end do
  !$omp end parallel

  return

end function calc_norm

!----------------------------------------------------------------------------------------------------!

function calc_acf_ccf(z,d,bigD,s,ovrlp,psi0)

  implicit none

  complex(kind=8),intent(in)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(in)::bigD(:)          !multi config amplitude (nconf)
  real(kind=8),intent(in)::s(:,:)              !action (nvb,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(in)::psi0(:)          !Initial wavepacket (dof)
  complex(kind=8),allocatable::temp_psi0(:)    !Temp variable for psi0 so it doesn't get changed
  complex(kind=8)::calc_acf_ccf
  integer::i,k

  allocate(temp_psi0(size(psi0)))
  
  calc_acf_ccf=(0.d0,0.d0)

  temp_psi0(:) = psi0(:)

  if (acf_ccf_flag.eq.2) temp_psi0(:)=-psi0(:) !If CCF, make initial wavepacket its mirror image
  
  !$omp parallel shared(temp_psi0,z,d,s,bigD) private(i,k) reduction(+:calc_acf_ccf)
  !$omp do collapse(2)
  do k=1,size(z,3) !nconf
    do i=1,size(z,2) !nbv (=1 for CCS)
      if (qsystem.eq.'2Q') then 
        calc_acf_ccf = calc_acf_ccf + exp(sum(conjg(temp_psi0(:))*z(:,i,k)-0.5d0*(conjg(temp_psi0(:))*temp_psi0(:)&
                                     +conjg(z(:,i,k))*z(:,i,k))))*d(i,k)*exp(im*s(i,k))*bigD(k)&
                                     *z(2,i,k)**real(n_prtcls)/sqrt(n_prtcls_fact)    !Extra bit due to second quantization
      else
        calc_acf_ccf = calc_acf_ccf + exp(sum(conjg(temp_psi0(:))*z(:,i,k)-0.5d0*(conjg(temp_psi0(:))*temp_psi0(:)&
                                    + conjg(z(:,i,k))*z(:,i,k))))*d(i,k)*exp(im*s(i,k))*bigD(k)
      end if
    end do
  end do
  !$omp end do
  !$omp end parallel

  deallocate(temp_psi0)

  return

end function calc_acf_ccf

!----------------------------------------------------------------------------------------------------!

subroutine calc_density_fock(z,bigD,s,density_fock,ovrlp)

  complex(kind=8),intent(in)::z(:,:,:)               !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::bigD(:)                !multi config amplitude (nconf)
  real(kind=8),intent(in)::s(:,:)                    !action (nvb,nconf)
  complex(kind=8),intent(inout)::density_fock(:,:)   !Density matrix (dof,dof)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)         !Overlap matrix (nbv,nbv,nconf,nconf)
  integer::m,n,k,l
 
  density_fock=(0.d0,0.d0)

 !$omp parallel shared(s,ovrlp,bigD,density_fock,z) private(l,k,n,m)
 !$omp do collapse(2)
 do n=1,size(density_fock,2)
   do m=1,size(density_fock,1)
     do l=1,size(bigD,1)
       do k=1,size(bigD,1)
          density_fock(m,n)=density_fock(m,n)+conjg(bigD(k))*bigD(l)*exp(im*(s(1,l)-s(1,k)))*ovrlp(1,1,k,l)&
                                             *conjg(z(m,1,k))*z(n,1,l)
        end do
      end do
    end do
  end do
  !$omp end do
  !$omp end parallel

  return

end subroutine calc_density_fock

!----------------------------------------------------------------------------------------------------!

function calc_density_1_body(x,density_fock)

  implicit none

  real(kind=8),intent(inout)::x                    !x coordinate for the 1 body density element you are calculating
  complex(kind=8),intent(in)::density_fock(:,:)    !Density indexed in terms of HO levels = <Psi | a* a |Psi>
  complex(kind=8)::calc_density_1_body             !1 body density, in terms of coordinate
  integer::m,n

  calc_density_1_body = cmplx(0.d0,0.d0)

  !$omp parallel shared(density_fock,x) private(n,m) reduction(+:calc_density_1_body)
  !$omp do collapse(2)
  do n=1,size(density_fock,2)
    do m=1,size(density_fock,1)
      calc_density_1_body = calc_density_1_body + 1/sqrt(2**(m-1)*calc_factorial(m-1))*(1/Pi)**0.25d0*exp(-(x**2.d0)/2.d0)&
                            *hermite(m-1,x)*density_fock(m,n)&
                            *1/sqrt(2**(n-1)*calc_factorial(n-1))*(1/Pi)**0.25d0*exp(-(x**2.d0)/2.d0)*hermite(n-1,x)
    end do
  end do
  !$omp end do
  !$omp end parallel

  return

end function calc_density_1_body

!----------------------------------------------------------------------------------------------------!

function calc_part_num(z,bigD,s,ovrlp)

  complex(kind=8),intent(in)::z(:,:,:)               !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::bigD(:)                !multi config amplitude (nconf)
  real(kind=8),intent(in)::s(:,:)                    !action (nvb,nconf)
  real(kind=8)::calc_part_num                     !Particle number in second quantized system <Psi |a*a| Psi>
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)         !Overlap matrix (nbv,nbv,nconf,nconf)
  integer::m,k,l
 
  calc_part_num=(0.d0,0.d0)

  
  if (qsystem.eq.'2Q') then
    !$omp parallel shared(s,bigD,z,ovrlp) private(l,k,m) reduction(+:calc_part_num)
    !$omp do collapse(3)
    do l=1,size(bigD,1)
      do k=1,size(bigD,1)
        do m=2,size(z,1)
          calc_part_num=calc_part_num+conjg(bigD(k))*bigD(l)*exp(im*(s(1,l)-s(1,k)))*ovrlp(1,1,k,l)&
                                             *conjg(z(m,1,k))*z(m,1,l)
        end do
      end do
    end do
    !$omp end do
    !$omp end parallel

  else if (qsystem.eq.'BC') then

    !$omp parallel shared(s,ovrlp,bigD,z) private(l,k,m) reduction(+:calc_part_num)
    !$omp do collapse(3)
    do l=1,size(bigD,1)
      do k=1,size(bigD,1)
        do m=1,size(z,1)
          calc_part_num=calc_part_num+conjg(bigD(k))*bigD(l)*exp(im*(s(1,l)-s(1,k)))*ovrlp(1,1,k,l)&
                                             *conjg(z(m,1,k))*z(m,1,l)
        end do
      end do
    end do
    !$omp end do
    !$omp end parallel

  end if

  return

end function calc_part_num

!----------------------------------------------------------------------------------------------------!

function calc_pos_variance(density_fock,normalised)

!Function to calculate <Q> = int dQ Q^2 rho(Q) - (int dQ Q rho(Q) )^2

 real(kind=8)::calc_pos_variance
 complex(kind=8),intent(in)::density_fock(:,:)   !Density matrix (dof,dof)
 logical,intent(in)::normalised !Whether the variance is normalised or not
 real(kind=8)::dens_trace !Trace of density matrix = particle number if normalised=.true., or =1 if normalised=.false.
 real(kind=8)::Q2_int,Q_int !The integrals Q2*rho(Q) and Q*rho(Q)
 integer::m,n
 

 calc_pos_variance=0.d0
 dens_trace=0.d0
 Q2_int=0.d0
 Q_int=0.d0

 if (normalised) then
   do m=1,size(density_fock,1)
     dens_trace = dens_trace + density_fock(m,m)
   end do
 else
   dens_trace=1.d0
 end if

 do n=1,size(density_fock,2)
   do m=1,size(density_fock,1)
     if (m.eq.n) then !diagonal term of <m|Q^2|n>, no analagous diagonal term for <m|Q|n>. 
       Q2_int = Q2_int + (density_fock(m,n)/dens_trace*0.5d0*(2.d0*(m-1.d0)+1.d0)) !Note relationship between counter and number of quanta 
     else if (m.eq.(n+1)) then !term <m|Q|n> = <m|Q|m-1> = <n+1|Q|n> 
       Q_int = Q_int + (density_fock(m,n)/dens_trace*sqrt(n*0.5d0))
     else if (n.eq.(m+1)) then !term <m|Q|n> = <m|Q|m+1> = <n-1|Q|n>
      Q_int = Q_int + (density_fock(m,n)/dens_trace*sqrt(m*0.5d0))
     else if (m.eq.(n+2)) then !term <m|Q^2|n> = <m|Q^2|m-2> = <n+2|Q^2|n>
       Q2_int = Q2_int + (density_fock(m,n)/dens_trace*0.5d0*sqrt((m-1.d0)*(m-2.d0))) 
     else if (n.eq.(m+2)) then !term <m|Q^2|n> = <m|Q^2|m+2> = <n-2|Q^2|n>
       Q2_int = Q2_int + (density_fock(m,n)/dens_trace*0.5d0*sqrt((n-1.d0)*(n-2.d0))) 
     end if
   end do
 end do

 calc_pos_variance = Q2_int - Q_int**2.d0

 return

end function calc_pos_variance

!----------------------------------------------------------------------------------------------------!


function hermite(level,x)
 
  implicit none

  integer,intent(in)::level          !Harmonic oscillator level you want to calculate the hermite polynomial for
  real(kind=8),intent(in)::x         !x coordinate for the hermite polynomal you are calculating
  real(kind=8)::hermite              !Value of the hermite polynomial for level alpha, coordinate x
  integer::pow_x

  hermite = 0.d0

  do pow_x=0,level
    hermite = hermite + hermite_coeff(pow_x+1,level+1)*x**real(pow_x,kind=8)      !pow_x+1 and level+1 because no 0 array index
  end do

  return

end function hermite

!----------------------------------------------------------------------------------------------------!

function calc_factorial(n)

  implicit none
  
  integer,intent(in)::n
  real(kind=8)::calc_factorial
  integer::n_counter

  calc_factorial = 1.d0

  do n_counter=1,n
    calc_factorial = calc_factorial*real(n_counter,kind=8) 
  end do

  return

end function calc_factorial

!----------------------------------------------------------------------------------------------------!

subroutine calc_bc_W_mnop

  implicit none
  
  integer::m,n,o,p
 
  bc_W_mnop(:,:,:,:) = 0.d0

  !$omp parallel shared(bc_W_mnop) private(m,n,o,p)
  !$omp do collapse(4)
  do p=1,size(bc_W_mnop,4) !dof
    do o=1,size(bc_W_mnop,3) !dof
      do n=1,size(bc_W_mnop,2) !dof
        do m=1,size(bc_W_mnop,1) !dof
          if (mod(((p-1)+(o-1)+(n-1)+(m-1)),2).eq.0) then !Only even overall functions will integrate out to nonzero matrix elements
            bc_W_mnop(m,n,o,p) = 1/Pi&
                                *1/sqrt(real(2.d0**(m-1),kind=8)*calc_factorial(m-1))& 
                                *1/sqrt(real(2.d0**(n-1),kind=8)*calc_factorial(n-1))&
                                *1/sqrt(real(2.d0**(o-1),kind=8)*calc_factorial(o-1))&
                                *1/sqrt(real(2.d0**(p-1),kind=8)*calc_factorial(p-1))&
                                *calc_hermite_int(m,n,o,p)
          end if
        end do
      end do
    end do
  end do
  !$omp end do
  !$omp end parallel

  return

end subroutine calc_bc_W_mnop

!----------------------------------------------------------------------------------------------------!


subroutine calc_hermite_coeff  !To calculate array of hermite coefficients

  implicit none

  integer::m,pow_x

  hermite_coeff(:,:) = 0.d0    !Set array initally to zero, then will only be changing non-zero elements

  do m=1,size(hermite_coeff,2) !order of Hermite
    do pow_x=m,1,-2 !Loop counter -2 as hermite function powers will either be all even or all odd
      if (pow_x.eq.m) then
        hermite_coeff(pow_x,m) = real(2.d0**(pow_x-1),kind=8)   !coefficient for largest power is 2^Hermite_order. Minus 1 because array index is one greater than power of x (can't have array index 0)
      else
        hermite_coeff(pow_x,m) = (pow_x+1.d0)*pow_x/(2.d0*(pow_x-m))*hermite_coeff(pow_x+2,m)  !Recurrance relation for generating hermite polynomial coefficients from the largest power
      end if
    end do
  end do

  return

end subroutine calc_hermite_coeff

!----------------------------------------------------------------------------------------------------!

function calc_hermite_int(m,n,o,p)

  integer,intent(in)::m,n,o,p                     !The order of the Hermite polynomials that are used for this calculation
  real(kind=8)::calc_hermite_int                  !Function to calculate the integral of the product of hermites
  real(kind=8),allocatable::hermite_prod_coeff(:) !Array for product of 4 Hermite polynomials (power of x)
  integer::pow_x_m,pow_x_n,pow_x_o,pow_x_p
  integer::high_pow_x,pow_x

  high_pow_x = (m-1)+(n-1)+(o-1)+(p-1)+1 !Highest power of x (+1) for this combination of Hermite polynomials (+-1 involved as array index = power+1)

  allocate(hermite_prod_coeff(high_pow_x))   !Allocate the array for the coefficients of the product of 4 Hermite polynomials to be the largest power of x possible for that combination

  hermite_prod_coeff(:) = 0.d0

  calc_hermite_int = 0.d0

  !Can't parallelise this, as this function is called from a parallel routine
  do pow_x_p=1,p
    do pow_x_o=1,o
      do pow_x_n=1,n
        do pow_x_m=1,m
          hermite_prod_coeff((pow_x_m-1)+(pow_x_n-1)+(pow_x_o-1)+(pow_x_p-1)+1)=&
          hermite_prod_coeff((pow_x_m-1)+(pow_x_n-1)+(pow_x_o-1)+(pow_x_p-1)+1)+hermite_coeff(pow_x_m,m)*hermite_coeff(pow_x_n,n)&
                                                                               *hermite_coeff(pow_x_o,o)*hermite_coeff(pow_x_p,p)
        end do
      end do
    end do 
  end do

  calc_hermite_int = hermite_prod_coeff(1)*sqrt(Pi/2.d0) !0 state

  do pow_x = (high_pow_x-1)/2,1,-1  !pow_x is actual power of x in prod of hermite polynomials, not array index
    calc_hermite_int = calc_hermite_int+sqrt(Pi/2.d0)/(real(4.d0**pow_x,kind=8))*calc_factorial(2*pow_x)&
                                           /real(2.d0**pow_x,kind=8)/calc_factorial(pow_x)&   !relationship between double factorial and single
                                            *hermite_prod_coeff(2*pow_x+1) !plus 1 because array starts at 1
  end do

  deallocate(hermite_prod_coeff)

  return
    
end function calc_hermite_int

!----------------------------------------------------------------------------------------------------!

end module auxiliary
