module basis

use constants; use auxiliary

!-----------------------------------------------------------------------------------------------!
!       Module to generate the basis set for CCS/2L-CCS simulation.                             !
!                                                                                               !
! Contains subroutines to:                                                                      !
! 1) Call routines to generate basis set elements                                               !
! 2) Generate coherent states with swarm sampling                                               !
! 3) Generate coherent states with pancake sampling                                             !
! 4) Generate single configurational amplitudes d                                               !
! 5) Generate multiconfigurational amplitudes bigD                                              !
!-----------------------------------------------------------------------------------------------!

contains

!-----------------------------------------------------------------------------------------------!

subroutine gen_basis(z,d,bigD,ovrlp,psi0)

  implicit none

  complex(kind=8),intent(inout)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(inout)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(inout)::bigD(:)          !multi config amplitude (nconf)
  complex(kind=8),intent(inout)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(inout)::psi0(:)          !Initial wavepacket (dof)

    select case (trim(sampling))
      case ("SWARM")
        call gen_z_swarm(z,psi0)
      case ("PANCAKE")
        call gen_z_pancake(z,psi0)
    end select

  call calc_ovrlp(ovrlp,z)

  call gen_d(d,z,ovrlp,psi0)
  call gen_bigD(bigD,z,d,ovrlp,psi0)

  return

end subroutine gen_basis

!-----------------------------------------------------------------------------------------------!

subroutine gen_z_swarm(z,psi0)

  implicit none

  complex(kind=8),intent(inout)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::psi0(:)             !initial position and momentum of wavepacket (dof)
  real(kind=8)::z_abs,phase_fact                  !Random number holders for second quantization sampling
  integer::i,k,m

  z(:,:,:)=cmplx(0.d0,0.d0) !Set all modes to 0 initially

  do k=1,size(z,3) !nconf
    do i=1,size(z,2) !nbv (=1 for CCS)
      do m=1,size(z,1) !dof
        if (qsystem.eq.'BC') then
          call random_number(phase_fact)
          if (m.eq.1) z_abs = ZBQLGAM(sys_cmprss*(n_prtcls+1.d0),sys_cmprss)
          if (m.ne.1) z_abs = ZBQLGAM(1.d0,cmprss)
          z(m,i,k) = cmplx(sqrt(abs(z_abs))*cos(phase_fact*2.d0*Pi),sqrt(abs(z_abs))*sin(phase_fact*2.d0*Pi),kind=8)
        else
          z(m,i,k) = cmplx((ZBQLNOR(real(psi0(m)),sigq/cmprss)),(ZBQLNOR(imag(psi0(m)),sigp/cmprss)),kind=8)
        end if
      end do
    end do
  end do

  return

end subroutine gen_z_swarm

!-----------------------------------------------------------------------------------------------!

subroutine gen_z_pancake(z,psi0)

  complex(kind=8),intent(inout)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::psi0(:)             !initial position and momentum of wavepacket (dof)
  real(kind=8)::z_abs,phase_fact                  !Random number holders for second quantization sampling like Harmonic Oscillator
  integer::i,k,m,rand_mode,uw_counter
  real(kind=8)::rand_num

  uw_counter=0 !set counter for number of basis functions placed in upper well (for AS and 2Q)

  z(:,:,:)=(0.d0,0.d0) !Set all modes to 0 for infinite compression

  do k=1,size(z,3) !nconf
    do i=1,size(z,2) !nbv (=1 for CCS)

      if (trim(method).eq."2LCCS") then !System part
        do m=1,sys_dof
          if ((m.eq.1).and.(as_upper_well_dim.ne.0).and.(uw_counter.lt.as_upper_well_dim)) then !For upper well basis function sampling
            z(m,i,k) = cmplx((ZBQLNOR(real(-psi0(m)),sigq/sys_cmprss)),(ZBQLNOR(imag(-psi0(m)),sigp/sys_cmprss)),kind=8)
            uw_counter=uw_counter+1
          else 
            z(m,i,k) = cmplx((ZBQLNOR(real(psi0(m)),sigq/sys_cmprss)),(ZBQLNOR(imag(psi0(m)),sigp/sys_cmprss)),kind=8)
          end if
        end do
      else
        if ((as_upper_well_dim.ne.0).and.(uw_counter.lt.as_upper_well_dim)) then !For upper well basis function sampling
          z(1,1,k) = cmplx((ZBQLNOR(real(-psi0(1)),sigq/sys_cmprss)),(ZBQLNOR(imag(-psi0(1)),sigp/sys_cmprss)),kind=8)
          uw_counter=uw_counter+1
        else 
          z(1,1,k) = cmplx((ZBQLNOR(real(psi0(1)),sigq/sys_cmprss)),(ZBQLNOR(imag(psi0(1)),sigp/sys_cmprss)),kind=8)
         end if
      end if

      if (i.eq.1) then !Bath part

        if (bath_sampling.eq.0) then !Normal pancake sampling 

          do m=(sys_dof+1),size(z,1) !dof (bv all the same for bath for given m and k in 2LCCS, and bv index is only 1 for CCS)
            if (qsystem.eq.'2Q') then !For the second quantized Hamiltonian, need to bath sample like a harmonic oscillator (same for CCS and 2LCCS)

              if (m.eq.2) then !If sys_dof=1 then this will be used
                call random_number(phase_fact)
                z_abs = ZBQLGAM(n_prtcls+1.d0,1.d0)
                z(m,:,k) = cmplx(sqrt(abs(z_abs))*cos(phase_fact*2.d0*Pi),sqrt(abs(z_abs))*sin(phase_fact*2.d0*Pi),kind=8)
              else
                call random_number(phase_fact)
                z_abs = ZBQLGAM(1.d0,cmprss)
                z(m,:,k) = cmplx(sqrt(abs(z_abs))*cos(phase_fact*2.d0*Pi),sqrt(abs(z_abs))*sin(phase_fact*2.d0*Pi),kind=8)
              end if

            else if ((qsystem.ne.'2Q').and.(cmprss.lt.1d10)) then      !For all other Hamiltonians, sample bath as normal (swarm) - second part of if statement is for infinite compression (z remains at 0)
                z(m,:,k) = cmplx((ZBQLNOR(real(psi0(m)),sigq/cmprss)),(ZBQLNOR(imag(psi0(m)),sigp/cmprss)),kind=8)
            end if
          end do

        else if (bath_sampling.eq.(dof-1)) then !Random variable linear decompression of all bath modes

          m=1 !Define m=1 to loop over all (dof-1) bath modes (currently will only work when sys_dof=1)

          do while (m.le.(dof-1)) !loop over all bath modes to be decompressed randomly
            call random_number(rand_num)
            rand_mode = int(rand_num*(dof-1))+2
            if (z(rand_mode,i,k).ne.(0.d0,0.d0)) then
              cycle  !if the random number is the same as previous (z already occupied) then cycle the loop to find another
            else
              z(rand_mode,:,k) = cmplx((ZBQLNOR(real(psi0(rand_mode)),sigq/(cmprss*m))),&
                                 (ZBQLNOR(imag(psi0(rand_mode)),sigp/(cmprss*m))),kind=8)
              m=m+1
            end if
          end do

        else if ((bath_sampling.gt.(dof-1)).or.(bath_sampling.lt.0)) then

          print*,"Unable to determine type of pancake sampling. Choose a bath_sampling value in the range [0:(dof-1)]"
          stop

        else !Decompression of bath_sampling number of bath modes, infinite compression on the rest

          m=1 !Define m=1 to loop over bath_sampling bath modes (currently will only work when sys_dof=1)

          do while (m.le.bath_sampling) !loop bath_sampling number of bath modes to be decompressed
            call random_number(rand_num)
            rand_mode = int(rand_num*(dof-1))+2
            if (z(rand_mode,i,k).ne.(0.d0,0.d0)) then
              cycle  !if the random number is the same as previous (z already occupied) then cycle the loop to find another
            else
              z(rand_mode,:,k) = cmplx((ZBQLNOR(real(psi0(rand_mode)),sigq/(cmprss))),&
                                 (ZBQLNOR(imag(psi0(rand_mode)),sigp/(cmprss))),kind=8)
              m=m+1
            end if
          end do
        end if
      end if
    end do
  end do

  return

end subroutine gen_z_pancake

!-----------------------------------------------------------------------------------------------!

subroutine gen_d(d,z,ovrlp,psi0)

  implicit none

  complex(kind=8),intent(inout)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(in)::z(:,:,:)            !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(inout)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(in)::psi0(:)             !Initial wavefunction
  complex(kind=8),allocatable::c(:)               !<z^q|Psi^q(0)> overlap for 2LCCS (nbv)
  integer::i,k

  !This routine currently only works for sys_dof=1, can be modified if needed

  if (trim(method).eq."CCS") then
    d(1,:) = (1.d0,0.d0)
  else if (trim(method).eq."2LCCS") then
    allocate(c(nbv))

    do k=1,size(d,2) !nconf
      !$omp parallel shared(c,z,psi0,d,k) private(i)
      !$omp do
      do i=1,size(d,1) !nbv
        c(i) = exp(sum(conjg(z(1:sys_dof,i,k))*psi0(1:sys_dof)-0.5d0*(conjg(z(1:sys_dof,i,k))*z(1:sys_dof,i,k)&
               +conjg(psi0(1:sys_dof))*psi0(1:sys_dof))))
      end do
      !$omp end do
      !$omp end parallel
      call lineq(ovrlp(:,:,k,k),c,d(:,k))

    end do

    deallocate(c)
  end if

  return

end subroutine gen_d

!-----------------------------------------------------------------------------------------------!

subroutine gen_bigD(bigD,z,d,ovrlp,psi0)

  implicit none

  complex(kind=8),intent(inout)::bigD(:)                !multi config amplitude (nconf)
  complex(kind=8),intent(in)::z(:,:,:)               !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)                 !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)         !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(in)::psi0(:)                !Initial wavepacket (dof)
  complex(kind=8),allocatable::C(:)                  !<z|Psi(0)> overlap (nconf)
  complex(kind=8),allocatable::LHS_bigD(:,:)            !Overlap*d (nconf,nconf)
  integer::i,j,k,l

  allocate(C(size(bigD)),LHS_bigD(size(bigD),size(bigD)))
  C(:) = (0.d0,0.d0)
  LHS_bigD(:,:) = (0.d0,0.d0)

  !$omp parallel shared(C,z,psi0) private(i,k)
  !$omp do
  do k=1,size(bigD) !nconf
    do i=1,size(z,2) !nbv(=1 for CCS)
      if (qsystem.eq.'2Q') then
        C(k) = C(k) + exp(sum(conjg(z(:,i,k))*psi0(:) - 0.5d0*(conjg(z(:,i,k))*z(:,i,k)+conjg(psi0(:))*psi0(:))))&
                      *conjg(z(2,i,k))**real(n_prtcls)/sqrt(n_prtcls_fact)
      else if (qsystem.eq.'BC') then
        C(k) = C(k) + exp(sum(-0.5d0*(conjg(z(:,i,k))*z(:,i,k))))&
                      *conjg(z(1,i,k))**real(n_prtcls)/sqrt(n_prtcls_fact)
      else
        C(k) = C(k) + exp(sum(conjg(z(:,i,k))*psi0(:) - 0.5d0*(conjg(z(:,i,k))*z(:,i,k)+conjg(psi0(:))*psi0(:))))
      end if
    end do
  end do
  !$omp end do
  !$omp end parallel

  !$omp parallel shared(LHS_bigD,d,ovrlp) private(i,j,k,l)
  !$omp do collapse(2)
  do l=1,size(ovrlp,4) !nconf
    do k=1,size(ovrlp,3) !nconf
      do j=1,size(ovrlp,2) !nbv (=1 for CCS)
        do i=1,size(ovrlp,1) !nbv (=1 for CCS)
          LHS_bigD(k,l) = LHS_bigD(k,l) + d(j,l)*ovrlp(i,j,k,l)
        end do
      end do
    end do
  end do
  !$omp end do
  !$omp end parallel

  call lineq(LHS_bigD,C,bigD)

  deallocate(C,LHS_bigD)

  return

end subroutine gen_bigD
      
!-----------------------------------------------------------------------------------------------!

end module basis
