module hamiltonian

use constants; 

!-----------------------------------------------------------------------------------------------!
!                           Hamiltonian module for CCS/2L-CCS simulation.                       !
!                                                                                               !
! Contains subroutines to:                                                                      !
! 1) Get the initial wavepacket psi0                                                            !
! 2) Calculate the normal ordered hamiltonian H                                                 !
! 3) Calculate the coherent state derivative of the normal ordered hamiltonian dHdz             !
!-----------------------------------------------------------------------------------------------!

contains

!-----------------------------------------------------------------------------------------------!

subroutine get_psi0(psi0)

  implicit none

  complex(kind=8),intent(inout)::psi0(:)      !Initial position and momentum of wavepacket for basis sampling (dof,nbv)

  select case (qsystem)

    case ('FP')
      psi0(:) = cmplx(5.d0*sigq,0.d0*sigp)
    case ('HP')
      psi0(:) = cmplx(5.d0*sigq,0.d0*sigp)
    case ('MP')
      psi0(:) = cmplx(5.d0*sigq,0.d0*sigp)
    case ('AS')
      psi0(1) = cmplx(-2.5d0*sigq,0.d0*sigp)
      psi0(2:dof) = cmplx(0.d0*sigq,0.d0*sigp)
    case ('2Q')
      psi0(1) = cmplx(-2.5d0*sigq,0.d0*sigp)
      psi0(2:dof) = cmplx(0.d0*sigq,0.d0*sigp)
    case ('BC')
      psi0(:) = (0.d0,0.d0)
  end select
  return

end subroutine get_psi0

!-----------------------------------------------------------------------------------------------!

subroutine calc_H(H,z)

  implicit none

  complex(kind=8),intent(inout)::H(:,:,:,:)       !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(in)::z(:,:,:)            !coherent state array (dof,nbv,nconf)
  integer::i,j,k,l,m,n,o,p

  H(:,:,:,:)=(0.d0,0.d0)

  select case (qsystem)

    case ('FP')

      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              do m=1,size(z,1) !dof
                H(i,j,k,l) = H(i,j,k,l) - 0.25d0*(conjg(z(m,i,k))**2.d0+z(m,j,l)**2.d0-2.d0*conjg(z(m,i,k))*z(m,j,l)-1.d0)
              end do
            end do
          end do
        end do
      end do

    case ('HP')

      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              do m=1,size(z,1) !dof
                H(i,j,k,l) = H(i,j,k,l) + conjg(z(m,i,k))*z(m,j,l)+0.5d0
              end do
            end do
          end do
        end do
      end do

    case ('MP')

      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              do m=1,size(z,1) !dof
                H(i,j,k,l) = H(i,j,k,l) - 0.25d0*(conjg(z(m,i,k))**2.d0+z(m,j,l)**2.d0-2.d0*conjg(z(m,i,k))*z(m,j,l)-1.d0)&
                             +mp_diss_en*(exp(-mp_well_param*sqrt(2.d0)*(conjg(z(m,i,k))+z(m,j,l))+mp_well_param**2.d0)&
                             -2.d0*exp(-mp_well_param/sqrt(2.d0)*(conjg(z(m,i,k))+z(m,j,l))+mp_well_param**2.d0/4.d0)+1.d0)
              end do
            end do
          end do
        end do
      end do

    case ('AS')

      !$omp parallel shared(H,z) private(m,i,j,k,l)
      !$omp do collapse(4)
      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              do m=2,size(z,1) !dof
                H(i,j,k,l) = H(i,j,k,l) + conjg(z(m,i,k))*z(m,j,l)+0.5d0+as_lambda/(4.d0*sqrt(2.d0))*(conjg(z(m,i,k))**2.d0&
                             +z(m,j,l)**2.d0+2.d0*conjg(z(m,i,k))*z(m,j,l)+1.d0)*(conjg(z(1,i,k))+z(1,j,l))
              end do
              H(i,j,k,l) = H(i,j,k,l) - 0.5d0*(conjg(z(1,i,k))**2.d0+z(1,j,l)**2.d0)+1.d0/(64.d0*as_eta)*(conjg(z(1,i,k))**4.d0&
                           +z(1,j,l)**4.d0+4.d0*conjg(z(1,i,k))**3.d0*z(1,j,l)+4.d0*conjg(z(1,i,k))*z(1,j,l)**3.d0&
                           +6.d0*conjg(z(1,i,k))**2.d0*z(1,j,l)**2.d0+12.d0*conjg(z(1,i,k))*z(1,j,l)&
                           +6.d0*conjg(z(1,i,k))**2.d0+6.d0*z(1,j,l)**2.d0+3.d0)
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

    case ('2Q')

      !$omp parallel shared(H,z) private(m,i,j,k,l)
      !$omp do collapse(4)
      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              H(i,j,k,l) = H(i,j,k,l) - 0.5d0*(conjg(z(1,i,k))**2.d0+z(1,j,l)**2.d0)+1.d0/(64.d0*as_eta)*(conjg(z(1,i,k))**4.d0&
                           +z(1,j,l)**4.d0+4.d0*conjg(z(1,i,k))**3.d0*z(1,j,l)+4.d0*conjg(z(1,i,k))*z(1,j,l)**3.d0&
                           +6.d0*conjg(z(1,i,k))**2.d0*z(1,j,l)**2.d0+12.d0*conjg(z(1,i,k))*z(1,j,l)&
                           +6.d0*conjg(z(1,i,k))**2.d0+6.d0*z(1,j,l)**2.d0+3.d0)
              do m=2,size(z,1) !dof
                H(i,j,k,l) = H(i,j,k,l) + (2.d0*m-3.5d0)*conjg(z(m,i,k))*z(m,j,l)&           !z z* eta [The relationship between the counter m and the number of quanta |k> is k=2(m-2)]
                             + as_lambda/2.d0*(conjg(z(1,i,k))+z(1,j,l))/sqrt(2.d0)&         !lambda*qs/2
                             *conjg(z(m,i,k))*z(m,j,l)*(2.d0*m-3.5d0)                            !Diagonal terms of Q_{k,k'}
                if (m.lt.size(z,1)) H(i,j,k,l) = H(i,j,k,l) + as_lambda/2.d0*(conjg(z(1,i,k))+z(1,j,l))/sqrt(2.d0)&
                                                              *(conjg(z((m+1),i,k))*z(m,j,l)+conjg(z(m,i,k))*z((m+1),j,l))&
                                                              *0.5d0*sqrt((2.d0*m-2.d0)*(2.d0*m-3.d0))         !Off diagonal terms of Q_{k,k'}
              end do
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

    case ('BC') 

      !$omp parallel shared(H,z) private(m,n,o,p,i,j,k,l)
      !$omp do collapse(4)
      do l=1,size(H,4)!nconf
        do k=1,size(H,3)!nconf
          do j=1,size(z,2)!nhbv (=1 for CCS)
            do i=1,size(z,2)!nhbv (=1 for CCS)
              do m=1,size(z,1)
                  H(i,j,k,l) = H(i,j,k,l) + conjg(z(m,i,k))*z(m,j,l)*(m-0.5d0) +conjg(z(m,i,k))*z(m,j,l)*2.205d0   !z z* eta [The relationship between the counter m and the number of quanta |k> is k=(m-1)]
                do n=1,size(z,1)
                  if (m.eq.(n+1)) then
                    H(i,j,k,l) = H(i,j,k,l) - 2.1d0*conjg(z(m,i,k))*z(n,j,l)*sqrt(real(n,kind=8)/2.d0)   !Not root{n+1) because of relationship between array index and number of quanta
                  else if (n.eq.(m+1)) then
                    H(i,j,k,l) = H(i,j,k,l) - 2.1d0*conjg(z(m,i,k))*z(n,j,l)*sqrt(real(m,kind=8)/2.d0) 
                  end if
                  do o=1,size(z,1)
                    do p=1,size(z,1)
                      H(i,j,k,l) = H(i,j,k,l)+conjg(z(m,i,k))*conjg(z(n,i,k))*z(o,j,l)*z(p,j,l)*bc_lambda*bc_W_mnop(m,n,o,p)*0.5d0 !Two body terms
                    end do
                  end do
                end do
              end do
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

  end select

  return

end subroutine calc_H

!-----------------------------------------------------------------------------------------------!

subroutine calc_dHdz(dHdz,z)

  complex(kind=8),intent(inout)::dHdz(:,:,:,:)    !Hamiltonian derivative (dof,nbv,nbv,nconf)
  complex(kind=8),intent(in)::z(:,:,:)            !coherent state array (dof,nbv,nconf)
  integer::i,j,k,m,n,o,p

  dHdz(:,:,:,:) = (0.d0,0.d0)

  select case (qsystem)

    case ('FP')

      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            do m=1,size(z,1) !dof
              dHdz(m,i,j,k) = -0.5d0*(conjg(z(m,i,k))-z(m,j,k)) 
            end do
          end do
        end do
      end do

    case ('HP')

      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            do m=1,size(z,1) !dof
              dHdz(m,i,j,k) = z(m,j,k) 
            end do
          end do
        end do
      end do

    case ('MP')

      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            do m=1,size(z,1) !dof
              dHdz(m,i,j,k) = -0.5d0*(conjg(z(m,i,k))-z(m,j,k))+sqrt(2.d0)*mp_diss_en*mp_well_param*&
                              (-exp(-mp_well_param*sqrt(2.d0)*(conjg(z(m,i,k))+z(m,j,k))+mp_well_param**2.d0)&
                               +exp(-mp_well_param/sqrt(2.d0)*(conjg(z(m,i,k))+z(m,j,k))+mp_well_param**2.d0/4.d0))
            end do
          end do
        end do
      end do

    case ('AS')

      !$omp parallel shared(dHdz,z) private(m,i,j,k)
      !$omp do collapse(3)
      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            do m=2,size(z,1) !dof
              dHdz(m,i,j,k) = z(m,j,k)+as_lambda/(2.d0*sqrt(2.d0))*(conjg(z(1,i,k))+z(1,j,k))*(conjg(z(m,i,k))+z(m,j,k))
            end do
            dHdz(1,i,j,k)=-conjg(z(1,i,k))+1.d0/(64.d0*as_eta)*(4.d0*conjg(z(1,i,k))**3.d0+12.d0*conjg(z(1,i,k))**2.d0*z(1,j,k)+&
                         4.d0*z(1,j,k)**3.d0+12.d0*conjg(z(1,i,k))*z(1,j,k)**2.d0+12.d0*z(1,j,k)+12.d0*conjg(z(1,i,k)))+sum(&
                         as_lambda/(4.d0*sqrt(2.d0))*(conjg(z(2:,i,k))**2.d0+z(2:,j,k)**2.d0+2.d0*conjg(z(2:,i,k))*z(2:,j,k)+1.d0))
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

    case ('2Q')
  
      !$omp parallel shared(dHdz,z) private(m,i,j,k)
      !$omp do collapse(3)
      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            dHdz(1,i,j,k) = dHdz(1,i,j,k) - conjg(z(1,i,k))+1.d0/(64.d0*as_eta)*(4.d0*conjg(z(1,i,k))**3.d0+&
                                            12.d0*conjg(z(1,i,k))**2.d0*z(1,j,k)+4.d0*z(1,j,k)**3.d0+&
                                            12.d0*conjg(z(1,i,k))*z(1,j,k)**2.d0+12.d0*z(1,j,k)+12.d0*conjg(z(1,i,k)))
            do m=2,size(z,1) !dof
              dHdz(1,i,j,k) = dHdz(1,i,j,k) + as_lambda/2.d0/sqrt(2.d0)*conjg(z(m,i,k))*z(m,j,k)*(2.d0*m-3.5d0)    !Diagonal terms of Q_{k,k'}

              dHdz(m,i,j,k) = dHdz(m,i,j,k) + (2.d0*m-3.5d0)*z(m,j,k)+as_lambda/2.d0*(conjg(z(1,i,k))+z(1,j,k))/sqrt(2.d0)&
                                              *z(m,j,k)*(2.d0*m-3.5d0)  !z eta and diagonal terms of Q_{k,k'}

              if (m.lt.size(z,1)) then
                dHdz(1,i,j,k) = dHdz(1,i,j,k) + as_lambda/2.d0/sqrt(2.d0)*(conjg(z((m+1),i,k))*z(m,j,k)&
                                                +conjg(z(m,i,k))*z((m+1),j,k))*0.5d0*sqrt((2.d0*m-2.d0)*(2.d0*m-3.d0)) !Off diagonal terms of Q_{k,k'} 

                dHdz(m,i,j,k) = dHdz(m,i,j,k) + as_lambda/2.d0*(conjg(z(1,i,k))+z(1,j,k))/sqrt(2.d0)&
                                                *z((m+1),j,k)*0.5d0*sqrt((2.d0*m-2.d0)*(2.d0*m-3.d0))   !Off diagonal terms of Q_{k,k'}  

                dHdz(m+1,i,j,k) = dHdz(m+1,i,j,k) + as_lambda/2.d0*(conjg(z(1,i,k))+z(1,j,k))/sqrt(2.d0)&
                                                    *z(m,j,k)*0.5d0*sqrt((2.d0*m-2.d0)*(2.d0*m-3.d0))   !Off diagonal terms of Q_{k,k'} for the next dof
              end if
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

    case ('BC')

      !$omp parallel shared(dHdz,z) private(m,n,o,p,i,j,k)
      !$omp do collapse(3)
      do k=1,size(z,3) !nconf
        do j=1,size(z,2) !nbv (=1 for CCS)
          do i=1,size(z,2) !nbv (=1 for CCS)
            do m=1,size(z,1)
                dHdz(m,i,j,k) = dHdz(m,i,j,k) + z(m,j,k)*(m-0.5d0)+z(m,j,k)*2.205d0        !z eta [The relationship between the counter m and the number of quanta |k> is k=(m-1)]
              do n=1,size(z,1)
                if (m.eq.(n+1)) then
                  dHdz(m,i,j,k) = dHdz(m,i,j,k) - 2.1d0*z(n,j,k)*sqrt(real(n,kind=8)/2.d0)
                else if (n.eq.(m+1)) then
                  dHdz(m,i,j,k) = dHdz(m,i,j,k) - 2.1d0*z(n,j,k)*sqrt(real(m,kind=8)/2.d0) 
                end if 
                do o=1,size(z,1)
                  do p=1,size(z,1) 
                    dHdz(m,i,j,k) = dHdz(m,i,j,k) + conjg(z(n,i,k))*z(o,j,k)*z(p,j,k)*bc_lambda*bc_W_mnop(m,n,o,p)*0.5d0  ! Two body terms
                    dHdz(n,i,j,k) = dHdz(n,i,j,k) + conjg(z(m,i,k))*z(o,j,k)*z(p,j,k)*bc_lambda*bc_W_mnop(m,n,o,p)*0.5d0     !Two body terms
                  end do
                end do
              end do
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

  end select

  return

end subroutine calc_dHdz

!-----------------------------------------------------------------------------------------------!

end module hamiltonian
