module initialise
use constants

!-----------------------------------------------------------------------------------------------!
!       Module to initialise CCS/2L-CCS simulation.                                             !
!                                                                                               !
! Contains subroutines to:                                                                      !
! 1) Read parameters from input file                                                            !
! 2) Allocate arrays                                                                            !
! 3) Generate seed for random number generator used in sampling of z                            !
! 4) Read basis from file                                                                       !
!-----------------------------------------------------------------------------------------------!

contains

!-----------------------------------------------------------------------------------------------!

subroutine readparams()

  implicit none
  character(len=100)::LINE,value
  integer::ierr,n

  open(130,file='../in/input.dat',status='old',iostat=ierr)

  do while (ierr.eq.0)

    read(130,*,iostat=ierr)LINE,value
    select case (trim(LINE))
      case ('dof')
        read(value,*)dof
      case ('sys_dof')
        read(value,*)sys_dof
      case ('nbv')
        read(value,*)nbv
      case ('nconf')
        read(value,*)nconf
      case ('dt')
        read(value,*)dt
      case ('time_end')
        read(value,*)time_end
      case ('cmprss')
        read(value,*)cmprss
      case ('sys_cmprss')
        read(value,*)sys_cmprss
      case ('method')
        read(value,*)method
      case ('qsystem')
        read(value,*)qsystem
      case ('sampling')
        read(value,*)sampling
      case ('as_upper_well_dim')
        read(value,*)as_upper_well_dim
      case ('bath_sampling')
        read(value,*)bath_sampling
      case ('bsout')
        read(value,*)bsout
      case ('readbs')
        read(value,*)readbs
      case ('reporting')
        read(value,*)reporting
      case ('n_prtcls')
        read(value,*)n_prtcls
    end select

  end do

  close(130)
 
  if (((qsystem.eq.'2Q').or.(qsystem.eq.'BC')).and.(n_prtcls.le.170)) then    !For second quantization, set n_prtcls_fact and make sure it doesn't go over 10^308
    n_prtcls_fact = 1.d0
    do n=1,n_prtcls
      n_prtcls_fact = n_prtcls_fact*n
    end do
  end if

  select case (qsystem)
    case('FP')
      max_norm=fp_max_norm
      min_norm=fp_min_norm
    case('HP')
      max_norm=hp_max_norm
      min_norm=hp_min_norm
    case('MP')
      max_norm=mp_max_norm
      min_norm=mp_min_norm
    case('AS','2Q')
      max_norm=as_max_norm
      min_norm=as_min_norm
    case('BC')
      max_norm=bc_max_norm
      min_norm=bc_min_norm
  end select

  return

end subroutine readparams

!-----------------------------------------------------------------------------------------------!

subroutine alarrays(z,d,bigD,s,H,ovrlp,psi0,density_fock)

  implicit none

  complex(kind=8),allocatable,intent(inout)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),allocatable,intent(inout)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),allocatable,intent(inout)::bigD(:)          !multi config amplitude (nconf)
  real(kind=8),allocatable,intent(inout)::s(:,:)              !action (nvb,nconf)
  complex(kind=8),allocatable,intent(inout)::H(:,:,:,:)       !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),allocatable,intent(inout)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf)
  complex(kind=8),allocatable,intent(inout)::psi0(:)          !Initial wavepacket (dof)
  complex(kind=8),allocatable,intent(inout)::density_fock(:,:)!Density matrix (dof,nconf)

  allocate(d(nbv,nconf),bigD(nconf),s(nbv,nconf),H(nbv,nbv,nconf,nconf),density_fock(dof,dof),bc_W_mnop(dof,dof,dof,dof)&
          ,hermite_coeff(dof,dof))

  if (trim(method).eq."CCS") then
    allocate(z(dof,1,nconf),ovrlp(1,1,nconf,nconf),psi0(dof))
  else if (trim(method).eq."2LCCS") then
    allocate(z(dof,nbv,nconf),ovrlp(nbv,nbv,nconf,nconf),psi0(dof))
  end if

  z(:,:,:) = (0.d0,0.d0)
  d(:,:) = (0.d0,0.d0)
  bigD(:) = (0.d0,0.d0)
  H(:,:,:,:) = (0.d0,0.d0)
  ovrlp(:,:,:,:) = (0.d0,0.d0)
  psi0(:) = 0.d0
  s(:,:) = 0.d0
  density_fock(:,:) = (0.d0,0.d0)
  bc_W_mnop(:,:,:,:) = 0.d0
  hermite_coeff(:,:) = 0.d0

  return

end subroutine alarrays

!-----------------------------------------------------------------------------------------------!

subroutine gen_seed()

  implicit none

  integer(kind=8)::ranseed
  integer,allocatable::ranseed_array(:)
  integer:: ranseed_size
  integer::istat,i,clock

  !This first part is for initialisation of the seed for the random_number() routine using the time for the seed

  call random_seed(size=ranseed_size)   
  allocate(ranseed_array(ranseed_size))

  call system_clock(count=clock)

  ranseed_array = clock + 37* (/ (i-1,i=1,ranseed_size) /)
  call random_seed(put=ranseed_array)

  deallocate(ranseed_array)

  !This next part is for initialisation of the ZBQLNOR seed

  open(unit=570, file="/dev/urandom", access="stream",form="unformatted", action="read", status="old", iostat=istat)

  if (istat == 0) then
     read(570) ranseed       ! This takes the random seed from the true-random bin. If the urandom bin does not
     close(570)              ! exist the random seed is set to zero which forces the date to be used
  else
    ranseed=0
  end if

  ranseed = abs(ranseed)    ! Negative seed values seem to cause instability

  call ZBQLINI(ranseed,0)   ! Generates the seed value using the UCL random library for ZBQLNOR routine

  return

end subroutine gen_seed

!-----------------------------------------------------------------------------------------------!

subroutine read_basis(t,z,s,d,bigD)

  implicit none

  real(kind=8),intent(inout)::t                   !time
  complex(kind=8),intent(inout)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(inout)::d(:,:)           !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(inout)::bigD(:)          !multi config amplitude (nconf)
  real(kind=8),intent(inout)::s(:,:)              !action (nvb,nconf)

  open(436,file="../out/outbs.out")
  read(436,'(E25.16)')t
  read(436,'(E25.16)')z
  read(436,'(E25.16)')s
  read(436,'(E25.16)')d
  read(436,'(E25.16)')bigD
  close(436)

  return

end subroutine read_basis

!-----------------------------------------------------------------------------------------------!
  

end module initialise

