module propagation

use constants; use auxiliary; use hamiltonian

!-----------------------------------------------------------------------------------------------!
! Propagation module to calculate time derivates of CSs and amplitudes for CCS/2L-CCS       !
! simulation and propagate via rk4 routine.                                                     !
!                                                                                               !
! Contains subroutines to:                                                                      !
! 1) Call seperate time derivatives routines                                                    !
! 2) Calculate coherent state time derivative dzdt                                              !
! 3) Calculate single configurational amplitude time derivative dddt                            !
! 4) Calculate time derivative of the action dsdt                                               !
! 5) Calculate multiconfigurational amplitude time derviative dbigDdt                           !
! 6) Propagate amplitudes and coherent states via rk4                                           !
!-----------------------------------------------------------------------------------------------!

contains

!-----------------------------------------------------------------------------------------------!

subroutine derivs(z,d,bigD,s,dzdt,dddt,dbigDdt,dsdt,ovrlp,H,dHdz)

  implicit none

  complex(kind=8),intent(inout)::z(:,:,:)              !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(inout)::d(:,:)                !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(inout)::bigD(:)               !multi config amplitude (nconf)
  real(kind=8),intent(inout)::s(:,:)                   !action (nvb,nconf)
  complex(kind=8),intent(inout)::dzdt(:,:,:)           !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),intent(inout)::dddt(:,:)             !Time derivative of d (nbv,nconf)
  complex(kind=8),intent(inout)::dbigDdt(:)            !Time derivateive of bigD (nconf)
  real(kind=8),intent(inout)::dsdt(:,:)                !Time derivative of s (nbv,nconf)
  complex(kind=8),intent(inout)::ovrlp(:,:,:,:)        !Overlap matrix (nbv,nbv,nconf,nconf) <z1|z2>
  complex(kind=8),intent(inout)::H(:,:,:,:)            !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),intent(inout)::dHdz(:,:,:,:)         !Hamiltonian derivative (dof,nbv,nbv,nconf)

  call calc_H(H,z)

  call calc_dHdz(dHdz,z)

  call calc_ovrlp(ovrlp,z)

  call calc_dzdt(dzdt,d,s,dHdz,ovrlp)

  call calc_dddt(dddt,dzdt,z,d,s,ovrlp,H)

  call calc_dsdt(dsdt,dzdt,z,H)

  call calc_dbigDdt(dbigDdt,dzdt,dddt,z,d,bigD,s,ovrlp,H)

  return

end subroutine derivs
!-----------------------------------------------------------------------------------------------!

subroutine calc_dzdt(dzdt,d,s,dHdz,ovrlp)

  implicit none

  complex(kind=8),intent(inout)::dzdt(:,:,:)   !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)           !factored single config amplitude (nbv,nconf)
  real(kind=8),intent(in)::s(:,:)              !action (nvb,nconf)
  complex(kind=8),intent(in)::dHdz(:,:,:,:)    !Hamiltonian derivative (dof,nbv,nbv,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)   !Overlap matrix (nbv,nbv,nconf,nconf) <z1|z2>
  complex(kind=8),allocatable::dzdt_bath(:,:)  !Time derivative of z_bath (dof,nconf) as bath states all have same value for given nbv index.
  integer::i,j,k,m

  dzdt(:,:,:) = (0.d0,0.d0)

  select case (trim(method))

    case ("CCS")

      !$omp parallel shared(dzdt,dHdz) private(k,m)
      !$omp do collapse(2)
      do k=1,size(dzdt,3) !nconf
        do m=1,size(dzdt,1) !dof
          dzdt(m,1,k) = -im*dHdz(m,1,1,k)            !nbv = 1 for CCS, so may as well explicitly declare this array dimension in the CCS case statement
        end do
      end do
      !$omp end do
      !$omp end parallel

    case ("2LCCS")

      allocate(dzdt_bath(size(dzdt,1),size(dzdt,3)))

      dzdt_bath(:,:) = (0.d0,0.d0)

      !$omp parallel shared(dzdt,dHdz,dzdt_bath,d,s,ovrlp) private(i,j,k,m)
      !$omp do collapse(2)
      do k=1,size(dzdt,3) !nconf
        do j=1,size(dzdt,2) !nbv
          do m=1,sys_dof
            dzdt(m,j,k) = -im*dHdz(m,j,j,k)             !Quantum/CCS part
          end do
          do i=1,size(dzdt,2) !nbv
            do m=(sys_dof+1),size(dzdt,1) !dof
              dzdt_bath(m,k) = dzdt_bath(m,k) - im*dHdz(m,i,j,k)*conjg(d(i,k))*d(j,k)*exp(im*(s(j,k)-s(i,k)))*ovrlp(i,j,k,k)    !Classical part, with all nbv indeces the same
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel
     
      !$omp parallel shared(dzdt,dzdt_bath) private(i)
      !$omp do
      do i=1,size(dzdt,2) !nbv
        dzdt((sys_dof+1):dof,i,:)=dzdt_bath((sys_dof+1):dof,:)
      end do
      !$omp end do
      !$omp end parallel
      deallocate(dzdt_bath)

  end select

 return

end subroutine calc_dzdt

!-----------------------------------------------------------------------------------------------!

subroutine calc_dddt(dddt,dzdt,z,d,s,ovrlp,H)

  implicit none
  
  complex(kind=8),intent(inout)::dddt(:,:)                        !Time derivative of d (nbv,nconf)
  complex(kind=8),intent(in)::dzdt(:,:,:)                         !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),intent(in)::z(:,:,:)                            !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)                              !factored single config amplitude (nbv,nconf)
  real(kind=8),intent(in)::s(:,:)                                 !action (nvb,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)                      !Overlap matrix (nbv,nbv,nconf,nconf) <z1|z2>
  complex(kind=8),intent(in)::H(:,:,:,:)                          !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),allocatable::LHS_d(:,:),RHS_d(:,:)              !Left and right hand side elements for the set of linear equations to find dddt for 2LCCS (nbv,nbv)
  complex(kind=8),allocatable::RHS_d_vec(:)                       !RHS vector of the linear equation (matrix multiplication of RHS_d and d)
  integer::i,j,k

  dddt(:,:) = (0.d0,0.d0) !For CCS, there is no d amplitude, so it doesn't change from original value (1), therefore dddt equals 0

  select case (trim(method))

    case("2LCCS")

      allocate(LHS_d(nbv,nbv),RHS_d(nbv,nbv),RHS_d_vec(nbv))

      do k=1,size(dddt,2) !nconf
        !$omp parallel shared(LHS_d,RHS_d,ovrlp,s,H,dzdt,z) private(i,j)
        !$omp do collapse(2)
        do j=1,size(dddt,1) !nbv
          do i=1,size(dddt,1) !nbv
            LHS_d(i,j) = ovrlp(i,j,k,k)*exp(im*s(j,k))
            RHS_d(i,j) = LHS_d(i,j)*im*(H(j,j,k,k)-H(i,j,k,k)+sum(im*dzdt(:,j,k)*(conjg(z(:,i,k))-conjg(z(:,j,k)))))
          end do
        end do
        !$omp end do
        !$omp end parallel

        RHS_d_vec = matmul(RHS_d,d(:,k))

        call lineq(LHS_d,RHS_d_vec,dddt(:,k))

      end do

      deallocate(LHS_d,RHS_d,RHS_d_vec)

  end select
      
  return

end subroutine calc_dddt

!-----------------------------------------------------------------------------------------------!

subroutine calc_dsdt(dsdt,dzdt,z,H)

  implicit none

  real(kind=8),intent(inout)::dsdt(:,:)        !Time derivative of s (nbv,nconf)
  complex(kind=8),intent(in)::dzdt(:,:,:)      !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),intent(in)::z(:,:,:)         !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::H(:,:,:,:)       !Hamiltonian (nbv,nbv,nconf,nconf)
  integer::i,k

  !$omp parallel shared(dsdt,z,dzdt,H) private (i,k)
  !$omp do collapse(2)
  do k=1,size(dsdt,2) !nconf
    do i=1,size(dsdt,1) !nbv (=1 for CCS)
      dsdt(i,k) = 0.5d0*im*sum(conjg(z(:,i,k))*dzdt(:,i,k)-conjg(dzdt(:,i,k))*z(:,i,k))-H(i,i,k,k)
    end do
  end do
  !$omp end do
  !$omp end parallel
  
  return

end subroutine calc_dsdt

!-----------------------------------------------------------------------------------------------!

subroutine calc_dbigDdt(dbigDdt,dzdt,dddt,z,d,bigD,s,ovrlp,H)

  implicit none

  complex(kind=8),intent(inout)::dbigDdt(:)                       !Time derivateive of bigD (nconf)
  complex(kind=8),intent(in)::dzdt(:,:,:)                         !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),intent(in)::dddt(:,:)                           !Time derivative of d (nbv,nconf)
  complex(kind=8),intent(in)::z(:,:,:)                            !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(in)::d(:,:)                              !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(in)::bigD(:)                             !multi config amplitude (nconf)
  real(kind=8),intent(in)::s(:,:)                                 !action (nvb,nconf)
  complex(kind=8),intent(in)::ovrlp(:,:,:,:)                      !Overlap matrix (nbv,nbv,nconf,nconf) <z1|z2>
  complex(kind=8),intent(in)::H(:,:,:,:)                          !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),allocatable::LHS_bigD(:,:),RHS_bigD(:,:)        !Left and right hand side elements of the set of linear equations for dbigDdt
  complex(kind=8),allocatable::RHS_bigD_vec(:)                    !RHS vector of the linear equation (matrix multiplication of RHS_bigD and bigD)
  integer::i,j,k,l

  allocate(LHS_bigD(size(dbigDdt),size(dbigDdt)),RHS_bigD(size(dbigDdt),size(dbigDdt)),RHS_bigD_vec(size(dbigDdt)))

  LHS_bigD(:,:) = (0.d0,0.d0)
  RHS_bigD(:,:) = (0.d0,0.d0)
  RHS_bigD_vec(:) = (0.d0,0.d0)
  dbigDdt(:) = (0.d0,0.d0)

  select case (trim(method))

    case("CCS")

      !$omp parallel shared(LHS_bigD,RHS_bigD,ovrlp,s,H,dzdt,z) private(k,l)
      !$omp do collapse(2)    
      do l=1,size(dbigDdt) !nconf
        do k=1,size(dbigDdt) !nconf
          LHS_bigD(k,l) = ovrlp(1,1,k,l)*exp(im*s(1,l))
          RHS_bigD(k,l) = LHS_bigD(k,l)*im*(H(1,1,l,l)-H(1,1,k,l)+sum(im*dzdt(:,1,l)*(conjg(z(:,1,k))-conjg(z(:,1,l)))))
        end do
      end do
      !$omp end do
      !$omp end parallel

      RHS_bigD_vec = matmul(RHS_bigD,bigD)

      call lineq(LHS_bigD,RHS_bigD_vec,dbigDdt)

    case("2LCCS")

      !$omp parallel shared(LHS_bigD,RHS_bigD,d,ovrlp,s,dddt,H,dzdt,z) private(i,j,k,l)
      !$omp do collapse(2)
      do l=1,size(dbigDdt) !nconf
        do k=1,size(dbigDdt) !nconf   
          do j=1,size(H,2) !nbv
            do i=1,size(H,1) !nbv
              LHS_bigD(k,l) = LHS_bigD(k,l) + conjg(d(i,k))*d(j,l)*ovrlp(i,j,k,l)*exp(im*(s(j,l)-s(i,k)))
              RHS_bigD(k,l) = RHS_bigD(k,l) - conjg(d(i,k))*ovrlp(i,j,k,l)*exp(im*(s(j,l)-s(i,k)))*(dddt(j,l)+im*d(j,l)*(H(i,j,k,l)&
                                        -H(j,j,l,l)-sum(im*dzdt(:,j,l)*(conjg(z(:,i,k))-conjg(z(:,j,l))))))
            end do
          end do
        end do
      end do
      !$omp end do
      !$omp end parallel

      RHS_bigD_vec = matmul(RHS_bigD,bigD)

      call lineq(LHS_bigD,RHS_bigD_vec,dbigDdt)

  end select

  deallocate(LHS_bigD,RHS_bigD,RHS_bigD_vec)

  return

end subroutine calc_dbigDdt

!-----------------------------------------------------------------------------------------------!

subroutine rk4_prop(z,d,bigD,s,ovrlp,H)

  implicit none

  complex(kind=8),intent(inout)::z(:,:,:)              !coherent state array (dof,nbv,nconf)
  complex(kind=8),intent(inout)::d(:,:)                !factored single config amplitude (nbv,nconf)
  complex(kind=8),intent(inout)::bigD(:)               !multi config amplitude (nconf)
  real(kind=8),intent(inout)::s(:,:)                   !action (nvb,nconf)
  complex(kind=8),intent(inout)::ovrlp(:,:,:,:)        !Overlap matrix (nbv,nbv,nconf,nconf) <z1|z2>
  complex(kind=8),intent(inout)::H(:,:,:,:)            !Hamiltonian (nbv,nbv,nconf,nconf)
  complex(kind=8),allocatable::dHdz(:,:,:,:)           !Hamiltonian derivative (dof,nbv,nbv,nconf)
  complex(kind=8),allocatable::dzdt(:,:,:)             !Time derivative of z (dof,nbv,nconf)
  complex(kind=8),allocatable::dddt(:,:)               !Time derivative of d (nbv,nconf)
  complex(kind=8),allocatable::dbigDdt(:)              !Time derivateive of bigD (nconf)
  real(kind=8),allocatable::dsdt(:,:)                  !Time derivative of s (nbv,nconf)
  complex(kind=8),allocatable::k_z(:,:,:,:)            !Runge-Kutta increments of z (dof,nbv,nconf,inc) where the first dim of inc = k1 etc...
  complex(kind=8),allocatable::k_d(:,:,:)              !Runge-Kutta increments of d (nbv,nconf,inc) where the first dim of inc = k1 etc...
  complex(kind=8),allocatable::k_bigD(:,:)             !Runge-Kutta increments of bigD (nconf,inc) where the first dim of inc = k1 etc...  
  real(kind=8),allocatable::k_s(:,:,:)                 !Runge-Kutta increments of s (nbv,nconf,inc) where the first dim of inc = k1 etc...
  complex(kind=8),allocatable::z_new(:,:,:)            !Runge-Kutta temp variable for coherent state array (dof,nbv,nconf)
  complex(kind=8),allocatable::d_new(:,:)              !Runge-Kutta temp variable for factored single config amplitude (nbv,nconf)
  complex(kind=8),allocatable::bigD_new(:)             !Runge-Kutta temp variable for multi config amplitude (nconf)
  real(kind=8),allocatable::s_new(:,:)                 !Runge-Kutta temp variable for action (nvb,nconf)

  allocate(dHdz(size(z,1),size(z,2),size(z,2),size(z,3)))
  allocate(dzdt(size(z,1),size(z,2),size(z,3)),dddt(size(d,1),size(d,2)),dbigDdt(size(bigD)),dsdt(size(s,1),size(s,2)))
  allocate(k_z(size(z,1),size(z,2),size(z,3),4),k_d(size(d,1),size(d,2),4),k_bigD(size(bigD),4),k_s(size(s,1),size(s,2),4))
  allocate(z_new(size(z,1),size(z,2),size(z,3)),d_new(size(d,1),size(d,2)),bigD_new(size(bigD)),s_new(size(s,1),size(s,2)))

  z_new(:,:,:) = z(:,:,:)
  d_new(:,:) = d(:,:)
  s_new(:,:) = s(:,:)
  bigD_new(:) = bigD(:)

  call derivs(z_new,d_new,bigD_new,s_new,dzdt,dddt,dbigDdt,dsdt,ovrlp,H,dHdz)

  k_z(:,:,:,1) = dzdt(:,:,:)
  k_d(:,:,1) = dddt(:,:)
  k_s(:,:,1) = dsdt(:,:)
  k_bigD(:,1) = dbigDdt(:)

  z_new(:,:,:) = z(:,:,:) + dt/2.d0*k_z(:,:,:,1)
  d_new(:,:) = d(:,:) + dt/2.d0*k_d(:,:,1)
  s_new(:,:) = s(:,:) + dt/2.d0*k_s(:,:,1)
  bigD_new(:) = bigD(:) + dt/2.d0*k_bigD(:,1)

  call derivs(z_new,d_new,bigD_new,s_new,dzdt,dddt,dbigDdt,dsdt,ovrlp,H,dHdz)

  k_z(:,:,:,2) = dzdt(:,:,:)
  k_d(:,:,2) = dddt(:,:)
  k_bigD(:,2) = dbigDdt(:)
  k_s(:,:,2) = dsdt(:,:)

  z_new(:,:,:) = z(:,:,:) + dt/2.d0*k_z(:,:,:,2)
  d_new(:,:) = d(:,:) + dt/2.d0*k_d(:,:,2)
  s_new(:,:) = s(:,:) + dt/2.d0*k_s(:,:,2)
  bigD_new(:) = bigD(:) + dt/2.d0*k_bigD(:,2)

  call derivs(z_new,d_new,bigD_new,s_new,dzdt,dddt,dbigDdt,dsdt,ovrlp,H,dHdz)

  k_z(:,:,:,3) = dzdt(:,:,:)
  k_d(:,:,3) = dddt(:,:)
  k_bigD(:,3) = dbigDdt(:)
  k_s(:,:,3) = dsdt(:,:)

  z_new(:,:,:) = z(:,:,:) + dt*k_z(:,:,:,3)
  d_new(:,:) = d(:,:) + dt*k_d(:,:,3)
  s_new(:,:) = s(:,:) + dt*k_s(:,:,3)
  bigD_new(:) = bigD(:) + dt*k_bigD(:,3)

  call derivs(z_new,d_new,bigD_new,s_new,dzdt,dddt,dbigDdt,dsdt,ovrlp,H,dHdz)

  k_z(:,:,:,4) = dzdt(:,:,:)
  k_d(:,:,4) = dddt(:,:)
  k_bigD(:,4) = dbigDdt(:)
  k_s(:,:,4) = dsdt(:,:)

  z(:,:,:) = z(:,:,:) + dt/6.d0*(k_z(:,:,:,1)+2.d0*k_z(:,:,:,2)+2.d0*k_z(:,:,:,3)+k_z(:,:,:,4))
  d(:,:) = d(:,:) + dt/6.d0*(k_d(:,:,1)+2.d0*k_d(:,:,2)+2.d0*k_d(:,:,3)+k_d(:,:,4))
  s(:,:) = s(:,:) + dt/6.d0*(k_s(:,:,1)+2.d0*k_s(:,:,2)+2.d0*k_s(:,:,3)+k_s(:,:,4))
  bigD(:) = bigD(:) + dt/6.d0*(k_bigD(:,1)+2.d0*k_bigD(:,2)+2.d0*k_bigD(:,3)+k_bigD(:,4))

  deallocate(dzdt,dddt,dsdt,dbigDdt,k_z,k_d,k_s,k_bigD,z_new,d_new,s_new,bigD_new,dHdz)

  return

end subroutine rk4_prop

!-----------------------------------------------------------------------------------------------!

end module propagation
